from typing import Optional, Dict, Any, Tuple

import numpy as np
import warnings
from ..base import BaseSimObj

IDEAL = "Ideal"
NOISY = "Noisy"
TWO_STAGE = "TwoStage"


class Battery(BaseSimObj):
    """This class models the behavior of a battery and battery management system (BMS).

    Args:
        capacity (float): Capacity of the battery [kWh]
        init_charge (float): Initial charge of the battery [kWh]
        max_power (float): Maximum charging rate of the battery [kW]
    """

    def __init__(self, capacity, init_charge, max_power):
        if init_charge > capacity:
            raise ValueError("Initial Charge cannot be greater than capacity.")
        self._capacity = capacity
        self._current_charge = init_charge
        self._init_charge = init_charge
        self._max_power = max_power
        self._current_charging_power = 0

    @property
    def _soc(self):
        """ Returns the state of charge of the battery as a percent."""
        return self._current_charge / self._capacity

    @property
    def max_charging_power(self):
        """ Returns the maximum charging power of the Battery."""
        return self._max_power

    @property
    def current_charging_power(self):
        """ Returns the current draw of the battery on the AC side."""
        return self._current_charging_power

    def charge(self, pilot, voltage, period):
        """ Method to "charge" the battery

        Args:
            pilot (float): Pilot signal passed to the battery. [A]
            voltage (float): AC voltage provided to the battery charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            float: actual charging rate of the battery. [A]

        Raises:
            ValueError: if voltage or period are <= 0.
        """
        if voltage <= 0:
            raise ValueError("Voltage must be greater than 0. Got {0}".format(voltage))
        if period <= 0:
            raise ValueError("period must be greater than 0. Got {0}".format(voltage))

        # Rate which would fill the battery in period minutes.
        rate_to_full = (self._capacity - self._current_charge) / (period / 60)

        charge_power = min([pilot * voltage / 1000, self._max_power, rate_to_full])
        self._current_charge += charge_power * (period / 60)
        self._current_charging_power = charge_power
        return charge_power * 1000 / voltage

    def reset(self, init_charge=None):
        """ Reset battery to initial state. If init_charge is not
        given (is None), the battery is reset to its initial charge
        on initialization.

        Args:
            init_charge (float): charge battery should be reset to. [acnsim units]

        Returns:
            None
        """
        if init_charge is None:
            self._current_charge = self._init_charge
        else:
            if init_charge > self._capacity:
                raise ValueError("Initial Charge cannot be greater than capacity.")
            self._current_charge = init_charge
        self._current_charging_power = 0

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict = {}
        nn_attr_lst = [
            "_max_power",
            "_current_charging_power",
            "_current_charge",
            "_capacity",
            "_init_charge",
        ]
        for attr in nn_attr_lst:
            attribute_dict[attr] = getattr(self, attr)
        return attribute_dict, context_dict

    @classmethod
    def _from_dict_helper(cls, out_obj, attribute_dict):
        out_obj._current_charging_power = attribute_dict["_current_charging_power"]
        out_obj._current_charge = attribute_dict["_current_charge"]

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(
            attribute_dict["_capacity"],
            attribute_dict["_init_charge"],
            attribute_dict["_max_power"],
        )
        cls._from_dict_helper(out_obj, attribute_dict)
        return out_obj, loaded_dict


class Linear2StageBattery(Battery):
    """ Extends Battery with a simple piecewise linear model of battery dynamics based on SoC.

    Battery model based on a piecewise linear approximation of battery behavior. The battery will charge at the
    minimum of max_rate and the pilot until it reaches _transition_soc. After this, the maximum charging rate of the
    battery will decrease linearly to 0 at 100% state of charge.

    For more info on model: https://www.sciencedirect.com/science/article/pii/S0378775316317396

    All public attributes are the same as Battery.

    Args:
        noise_level (float): Standard deviation of the noise to add to the charging process. (kW)
        transition_soc (float): State of charging when transitioning from constant current to constraint voltage.
        charge_calculation (str): If 'stepwise', use the charging
            method from a previous version of acnportal, which
            assumes a constant maximal charging rate for the entire
            timestep during which the pilot signal is input. This
            charging method is less accurate than the _charge method,
            and should only be used for reproducing results from
            older versions of acnportal.

            If 'continuous' or not provided, use the _charge method,
            which assumes a continuously varying maximal charging rate.
    """

    charging_methods = ["continuous", "stepwise"]

    def __init__(
        self,
        capacity,
        init_charge,
        max_power,
        noise_level=0,
        transition_soc=0.8,
        charge_calculation="continuous",
    ):
        super().__init__(capacity, init_charge, max_power)
        self._noise_level = noise_level
        if transition_soc < 0:
            raise ValueError(
                f"transition_soc must be non-negative. " f"Got {transition_soc}."
            )
        elif transition_soc >= 1:
            raise ValueError(
                f"transition_soc must be less than 1. " f"Got {transition_soc}."
            )
        self._transition_soc = transition_soc
        if charge_calculation not in self.charging_methods:
            raise ValueError(
                f"Charging method {charge_calculation} specified in "
                f"charge_calculation attribute not recognized by "
                f"Linear2StageBattery class. use one of "
                f"{self.charging_methods}."
            )
        self.charge_calculation = charge_calculation

    def charge(self, pilot, voltage, period):
        """ Method to "charge" the battery based on a two-stage linear
        battery model.

        Uses one of
        {_charge, _charge_stepwise}
        to charge the battery depending on the value of the
        charge_calculation attribute of this object.
        """
        if self.charge_calculation == "stepwise":
            return self._charge_stepwise(pilot, voltage, period)
        elif self.charge_calculation == "continuous":
            return self._charge(pilot, voltage, period)
        else:
            raise ValueError(
                f"Charging method {self.charge_calculation} specified "
                f"in charge_calculation attribute not recognized by "
                f"Linear2StageBattery class. use one of "
                f"{self.charging_methods}."
            )

    def _charge(self, pilot, voltage, period):
        """ Method to "charge" the battery based on a two-stage linear
        battery model.

        All calculations are done in terms fo battery state of charge
        (SoC). Results are converted back to power units at the end.

        Args:
            pilot (float): Pilot signal passed to the battery. [A]
            voltage (float): AC voltage provided to the battery
                charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            float: average charging rate of the battery over this single
                period.

        """
        if voltage <= 0:
            raise ValueError(f"Voltage must be greater than 0. Got {voltage}.")
        if period <= 0:
            raise ValueError(f"Period must be greater than 0. Got {period}.")
        if pilot == 0:
            self._current_charging_power = 0
            return 0
        # All calculations are done in terms of battery SoC, so we
        # convert pilot signal and max power into pilot and max rate of
        # change of SoC.
        pilot_dsoc = pilot * voltage / 1000 / self._capacity / (60 / period)
        max_dsoc = self._max_power / self._capacity / (60 / period)

        if pilot_dsoc > max_dsoc:
            pilot_dsoc = max_dsoc

        # The pilot SoC rate of change has a new transition SoC at
        # which decreasing of max charging rate occurs.
        pilot_transition_soc = self._transition_soc + (
            pilot_dsoc - max_dsoc
        ) / max_dsoc * (self._transition_soc - 1)

        if pilot < 0:
            warnings.warn(
                f"Negative pilot signal input. Battery models"
                f"may not be accurate for pilot {pilot} A."
            )

        # The charging equation depends on whether the current SoC of
        # the battery is above or below the new transition SoC.
        if self._soc < pilot_transition_soc:
            # In the pre-rampdown region, the charging equation changes
            # depending on whether charging the battery over this
            # time period causes the battery to transition between
            # charging regions.
            if 1 <= (pilot_transition_soc - self._soc) / pilot_dsoc:
                curr_soc = pilot_dsoc + self._soc
            else:
                curr_soc = 1 + np.exp(
                    (pilot_dsoc + self._soc - pilot_transition_soc)
                    / (pilot_transition_soc - 1)
                ) * (pilot_transition_soc - 1)
        else:
            curr_soc = 1 + np.exp(pilot_dsoc / (pilot_transition_soc - 1)) * (
                self._soc - 1
            )

        # Add subtractive noise to the final SoC, scaling the noise
        # such that _noise_level is the standard deviation of the noise
        # in the battery charging power.
        if self._noise_level > 0:
            raw_noise = np.random.normal(0, self._noise_level)
            scaled_noise = raw_noise * (period / 60) / self._capacity
            curr_soc -= abs(scaled_noise)

        dsoc = curr_soc - self._soc
        self._current_charge = curr_soc * self._capacity

        # For charging power and charging rate (current), we use the
        # the average over this time period.
        self._current_charging_power = dsoc * self._capacity / (period / 60)
        return self._current_charging_power * 1000 / voltage

    def _charge_stepwise(self, pilot, voltage, period):
        """ Method to "charge" the battery based on a two-stage linear
        battery model.

        This is a legacy charging method from an older version of
        acnportal. This method assumes a constant maximal charging rate
        for the entire timestep during which the pilot signal is input.
        This charging method is less accurate than the _charge method,
        and should only be used for reproducing results from  older
        versions of acnportal.

        Args:
            pilot (float): Pilot signal passed to the battery. [A]
            voltage (float): AC voltage provided to the battery charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            float: actual charging rate of the battery.

        """
        if voltage <= 0:
            raise ValueError("Voltage must be greater than 0. Got {0}".format(voltage))
        if period <= 0:
            raise ValueError("period must be greater than 0. Got {0}".format(voltage))

        # Rate which would fill the battery in period minutes.
        rate_to_full = (self._capacity - self._current_charge) / (period / 60)

        if self._soc < self._transition_soc:
            charge_power = min([pilot * voltage / 1000, self._max_power, rate_to_full])
            if self._noise_level > 0:
                charge_power = max(
                    charge_power - abs(np.random.normal(0, self._noise_level)), 0
                )
        else:
            charge_power = min(
                [
                    pilot * voltage / 1000,
                    (1 - self._soc) / (1 - self._transition_soc) * self._max_power,
                    rate_to_full,
                ]
            )
            if self._noise_level > 0:
                charge_power = min(
                    max(charge_power + np.random.normal(0, self._noise_level), 0),
                    pilot * voltage / 1000,
                    self._max_power,
                    rate_to_full,
                )
                # ensure that noise does not cause the battery to violate any hard limits.
                charge_power = min(
                    [
                        charge_power,
                        pilot * voltage / 1000,
                        self._max_power,
                        rate_to_full,
                    ]
                )
        self._current_charge += charge_power * (period / 60)
        self._current_charging_power = charge_power
        return charge_power * 1000 / voltage

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict, context_dict = super()._to_dict(context_dict)
        attribute_dict["_noise_level"] = self._noise_level
        attribute_dict["_transition_soc"] = self._transition_soc
        attribute_dict["charge_calculation"] = self.charge_calculation
        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(
            attribute_dict["_capacity"],
            attribute_dict["_init_charge"],
            attribute_dict["_max_power"],
            noise_level=attribute_dict["_noise_level"],
            transition_soc=attribute_dict["_transition_soc"],
        )
        if "charge_calculation" in attribute_dict:
            out_obj.charge_calculation = attribute_dict["charge_calculation"]
        cls._from_dict_helper(out_obj, attribute_dict)
        return out_obj, loaded_dict


def batt_cap_fn(requested_energy, stay_dur, voltage, period):
    """ This function takes as input a requested energy, stay duration,
    and measurement parameters (voltage & period) and calculates the
    minimum capacity linear 2 stage battery such that it is
    feasible to deliver requested_energy in stay_dur periods.

    The function returns this minimum total capacity along with an
    initial capacity, which is the maximum initial capacity the battery
    with given total capacity may have such that it is feasible (after
    charging at max rate) to deliver requested_energy in stay_dur
    periods. Thus, the returned total and initial capacities maximize
    the amount of time the battery behaves non-ideally during charging.

    Args:
        requested_energy (float): Energy requested by this EV. If this
            fit uses data from ACN-Data, this requested_energy is the
            amount of energy actually delivered in real life.
        stay_dur (float): Number of periods the EV stayed.
        voltage (float): Voltage at which the battery is charged (V).
        period (float): Number of minutes in a period (minutes).

    """

    def _get_init_cap(battery_cap, max_rate=32, transition_soc=0.8):
        """ Given a requested energy, stay duration, battery capacity
        (kWh), charging voltage, period, max charging rate, and the
        state of charge (SoC) at which non-ideal behavior begins, finds
        the maximum initial capacity the battery may have such that
        requested_energy is delivered in stay_dur periods, assuming
        linear 2 stage battery behavior.

        """
        delta_soc = requested_energy / battery_cap
        # Maximum rate of change of SoC in per period.
        max_dsoc = max_rate * voltage / 1000 / battery_cap / (60 / period)

        # First, assume init_soc is after the transition_soc. In this
        # case, the max init_soc has a closed form solution.
        init_soc = 1 + delta_soc / (
            np.exp(max_dsoc * stay_dur / (transition_soc - 1)) - 1
        )
        if init_soc >= transition_soc:
            return init_soc

        # If that didn't work, search over all possible init_soc for the
        # largest init_soc that still allows for delta_soc to be
        # delivered in stay_dur periods.
        def delta_soc_from_init_soc(init_soc_guess):
            if stay_dur <= (transition_soc - init_soc_guess) / max_dsoc:
                return max_dsoc * stay_dur
            return (
                1
                + np.exp(
                    (max_dsoc * stay_dur + init_soc_guess - transition_soc)
                    / (transition_soc - 1)
                )
                * (transition_soc - 1)
                - init_soc_guess
            )

        # Before that, we make sure that starting at init_soc of 0, it's
        # possible to deliver the requested energy with this battery.
        if delta_soc_from_init_soc(0) < delta_soc:
            return -1

        # Since max energy delivered is decreasing in init_soc, we
        # use a binary search for a decreasing function. The default
        # tolerance is set to 1e-9 to satisfy the tests'
        # assertAlmostEqual default tolerance.
        def binsearch(f, lb, ub, target, tol=1e-9):
            mid = (lb + ub) / 2
            val = f(mid)
            if abs(val - target) < tol:
                return mid
            elif val - target > 0:
                return binsearch(f, mid, ub, target, tol=tol)
            else:
                return binsearch(f, lb, mid, target, tol=tol)

        # Below an initial capacity of
        # (transition_soc - max_dsoc * stay_dur),
        # the same amount of charge is delivered, as charging stays
        # entirely within the ideal region. So, we can start the binary
        # search here.
        init_soc = binsearch(
            delta_soc_from_init_soc, transition_soc - max_dsoc * stay_dur, 1, delta_soc
        )
        return init_soc * battery_cap

    potential_caps = np.array([8, 24, 40, 60, 85, 100])
    for cap in potential_caps:
        if requested_energy > cap:
            continue
        init = _get_init_cap(cap)
        if init >= 0:
            return cap, init
    raise ValueError("No feasible battery size found.")



""" Unit tests for analysis functions. """
from datetime import datetime
from unittest import TestCase
from unittest.mock import Mock, create_autospec

import numpy as np
import pytz

from acnportal import acnsim
from acnportal.algorithms import BaseAlgorithm


class TestAnalysis(TestCase):
    def setUp(self):
        start = datetime(2018, 12, 31, tzinfo=pytz.timezone("America/Los_Angeles"))
        network = acnsim.ChargingNetwork()
        evse1 = acnsim.EVSE("PS-001", max_rate=32)
        network.register_evse(evse1, 240, 0)
        scheduler = create_autospec(BaseAlgorithm)
        scheduler.max_recompute = None
        self.events = acnsim.EventQueue(events=[acnsim.Event(1)])
        self.simulator = acnsim.Simulator(
            network, scheduler, self.events, start, period=240
        )
        self.simulator._iteration = 10
        self.expected_datetime_array = [
            np.datetime64("2018-12-31T00:00:00.000000"),
            np.datetime64("2018-12-31T04:00:00.000000"),
            np.datetime64("2018-12-31T08:00:00.000000"),
            np.datetime64("2018-12-31T12:00:00.000000"),
            np.datetime64("2018-12-31T16:00:00.000000"),
            np.datetime64("2018-12-31T20:00:00.000000"),
            np.datetime64("2019-01-01T00:00:00.000000"),
            np.datetime64("2019-01-01T04:00:00.000000"),
            np.datetime64("2019-01-01T08:00:00.000000"),
            np.datetime64("2019-01-01T12:00:00.000000"),
        ]

    def test_datetimes_array_warning(self):
        with self.assertWarns(UserWarning):
            datetime_array = acnsim.datetimes_array(self.simulator)
            np.testing.assert_equal(datetime_array, self.expected_datetime_array)

    def test_datetimes_array(self):
        self.events.empty = Mock(self.events.empty)
        self.events.empty = lambda: True
        datetime_array = acnsim.datetimes_array(self.simulator)
        # Check that simulator start is unchanged.
        self.assertIsNotNone(self.simulator.start.tzinfo)
        np.testing.assert_equal(datetime_array, self.expected_datetime_array)



import copy
from datetime import datetime
from typing import Dict

import warnings
import json

# noinspection PyProtectedMember
from pydoc import locate

from .events import *
from .events import UnplugEvent
from .interface import Interface
from .interface import InvalidScheduleError
from acnportal.algorithms import BaseAlgorithm
from .base import BaseSimObj


class Simulator(BaseSimObj):
    """ Central class of the acnsim package.

    The Simulator class is the central place where everything about a particular simulation is stored including the
    network, scheduling algorithm, and events. It is also where timekeeping is done and orchestrates calling the
    scheduling algorithm, sending pilots to the network, and updating the energy delivered to each EV.

    Args:
        network (ChargingNetwork): The charging network which the simulation will use.
        scheduler (BaseAlgorithm): The scheduling algorithm used in the simulation.
            If scheduler = None, Simulator.run() cannot be called.
        events (EventQueue): Queue of events which will occur in the simulation.
        start (datetime): Date and time of the first period of the simulation.
        period (float): Length of each time interval in the simulation in minutes. Default: 1
        signals (Dict[str, ...]):
        store_schedule_history (bool): If True, store the scheduler output each time it is run. Note this can use lots
            of memory for long simulations.
        interface_type (type): The class of interface to register with the scheduler.
    """

    period: float

    def __init__(
        self,
        network,
        scheduler,
        events,
        start,
        period: float = 1,
        signals=None,
        store_schedule_history=False,
        verbose=True,
        interface_type=Interface,
    ):
        self.network = network
        self.scheduler = scheduler
        self.max_recompute = None
        self.event_queue = events
        self.start = start
        self.period = period
        self.signals = signals
        self.verbose = verbose

        # Information storage
        width = 1
        if self.event_queue.get_last_timestamp() is not None:
            width = self.event_queue.get_last_timestamp() + 1
        self.pilot_signals = np.zeros((len(self.network.station_ids), width))
        self.charging_rates = np.zeros((len(self.network.station_ids), width))
        self.peak = 0
        self.ev_history = {}
        self.event_history = []
        if store_schedule_history:
            self.schedule_history = {}
        else:
            self.schedule_history = None

        # Local Variables
        self._iteration = 0
        self._resolve = False
        self._last_schedule_update = None

        # Interface registration is moved here so that copies of this
        # simulator have all attributes.
        if scheduler is not None:
            self.max_recompute = scheduler.max_recompute
            self.scheduler.register_interface(interface_type(self))

    @property
    def iteration(self):
        return self._iteration

    def run(self):
        """
        If scheduler is not None, run the simulation until the event queue is empty.

        The run function is the heart of the simulator. It triggers all actions and keeps the simulator moving forward.
        Its actions are (in order):
            1. Get current events from the event queue and execute them.
            2. If necessary run the scheduling algorithm.
            3. Send pilot signals to the network.
            4. Receive back actual charging rates from the network and store the results.

        Returns:
            None

        Raises:
            TypeError: If called when the scheduler attribute is None.
                The run() method requires a BaseAlgorithm-like
                scheduler to execute.
        """
        if self.scheduler is None:
            raise TypeError("Add a scheduler before attempting to call" " run().")
        while not self.event_queue.empty():
            current_events = self.event_queue.get_current_events(self._iteration)
            for e in current_events:
                self.event_history.append(e)
                self._process_event(e)
            if (
                self._resolve
                or self.max_recompute is not None
                and (
                    self._last_schedule_update is None
                    or self._iteration - self._last_schedule_update
                    >= self.max_recompute
                )
            ):
                new_schedule = self.scheduler.run()
                self._update_schedules(new_schedule)
                if self.schedule_history is not None:
                    self.schedule_history[self._iteration] = new_schedule
                self._last_schedule_update = self._iteration
                self._resolve = False
            if not self.event_queue.empty():
                width_increase = self.event_queue.get_last_timestamp() + 1
            else:
                width_increase = self._iteration + 1
            self.pilot_signals = _increase_width(self.pilot_signals, width_increase)
            self.charging_rates = _increase_width(self.charging_rates, width_increase)
            self.network.update_pilots(self.pilot_signals, self._iteration, self.period)
            self._store_actual_charging_rates()
            self.network.post_charging_update()
            self._iteration = self._iteration + 1

    def step(self, new_schedule):
        """ Step the simulation until the next schedule recompute is
        required.

        The step function executes a single iteration of the run()
        function. However, the step function updates the simulator with
        an input schedule rather than query the scheduler for a new
        schedule when one is required. Also, step will return a flag if
        the simulation is done.

        Args:
            new_schedule (Dict[str, List[number]]): Dictionary mapping
                station ids to a schedule of pilot signals.

        Returns:
            bool: True if the simulation is complete.
        """
        while (
            not self.event_queue.empty()
            and not self._resolve
            and (
                self.max_recompute is None
                or (self._iteration - self._last_schedule_update < self.max_recompute)
            )
        ):
            self._update_schedules(new_schedule)
            if self.schedule_history is not None:
                self.schedule_history[self._iteration] = new_schedule
            self._last_schedule_update = self._iteration
            self._resolve = False
            if self.event_queue.get_last_timestamp() is not None:
                width_increase = max(
                    self.event_queue.get_last_timestamp() + 1, self._iteration + 1
                )
            else:
                width_increase = self._iteration + 1
            self.pilot_signals = _increase_width(self.pilot_signals, width_increase)
            self.charging_rates = _increase_width(self.charging_rates, width_increase)
            self.network.update_pilots(self.pilot_signals, self._iteration, self.period)
            self._store_actual_charging_rates()
            self.network.post_charging_update()
            self._iteration = self._iteration + 1
            current_events = self.event_queue.get_current_events(self._iteration)
            for e in current_events:
                self.event_history.append(e)
                self._process_event(e)
        return self.event_queue.empty()

    def get_active_evs(self):
        """ Return all EVs which are plugged in and not fully charged at the current time.

        Wrapper for self.network.active_evs. See its documentation for more details.

        Returns:
            List[EV]: List of all EVs which are plugged in but not fully charged at the current time.

        """
        evs = copy.deepcopy(self.network.active_evs)
        return evs

    def _process_event(self, event):
        """ Process an event and take appropriate actions.

        Args:
            event (Event): Event to be processed.

        Returns:
            None
        """
        if event.event_type == "Plugin":
            self._print("Plugin Event...")
            self.network.plugin(event.ev)
            self.ev_history[event.ev.session_id] = event.ev
            self.event_queue.add_event(UnplugEvent(event.ev.departure, event.ev))
            self._resolve = True
            self._last_schedule_update = event.timestamp
        elif event.event_type == "Unplug":
            self._print("Unplug Event...")
            self.network.unplug(event.ev.station_id, event.ev.session_id)
            self._resolve = True
            self._last_schedule_update = event.timestamp
        elif event.event_type == "Recompute":
            self._print("Recompute Event...")
            self._resolve = True

    def _update_schedules(self, new_schedule):
        """ Extend the current self.pilot_signals with the new pilot signal schedule.

        Args:
            new_schedule (Dict[str, List[number]]): Dictionary mapping station ids to a schedule of pilot signals.

        Returns:
            None

        Raises:
            KeyError: Raised when station_id is in the new_schedule but not registered in the Network.
        """
        if len(new_schedule) == 0:
            return

        for station_id in new_schedule:
            if station_id not in self.network.station_ids:
                raise KeyError(
                    "Station {0} in schedule but not found in network.".format(
                        station_id
                    )
                )

        schedule_lengths = set(len(x) for x in new_schedule.values())
        if len(schedule_lengths) > 1:
            raise InvalidScheduleError("All schedules should have the same length.")
        schedule_length = schedule_lengths.pop()

        schedule_matrix = np.array(
            [
                new_schedule[evse_id]
                if evse_id in new_schedule
                else [0] * schedule_length
                for evse_id in self.network.station_ids
            ]
        )
        if not self.network.is_feasible(schedule_matrix):
            aggregate_currents = self.network.constraint_current(schedule_matrix)
            diff_vec = (
                np.abs(aggregate_currents)
                - np.tile(
                    self.network.magnitudes + self.network.violation_tolerance,
                    (schedule_length, 1),
                ).T
            )
            max_idx = np.unravel_index(np.argmax(diff_vec), diff_vec.shape)
            max_diff = diff_vec[max_idx]
            max_timeidx = max_idx[1]
            max_constraint = self.network.constraint_index[max_idx[0]]
            warnings.warn(
                f"Invalid schedule provided at iteration {self._iteration}. "
                f"Max violation is {max_diff} A on {max_constraint} "
                f"at time index {max_timeidx}.",
                UserWarning,
            )
        if self._iteration + schedule_length <= self.pilot_signals.shape[1]:
            self.pilot_signals[
                :, self._iteration : (self._iteration + schedule_length)
            ] = schedule_matrix
        else:
            # We've reached the end of pilot_signals, so double pilot_signal array width
            self.pilot_signals = _increase_width(
                self.pilot_signals,
                max(
                    self.event_queue.get_last_timestamp() + 1,
                    self._iteration + schedule_length,
                ),
            )
            self.pilot_signals[
                :, self._iteration : (self._iteration + schedule_length)
            ] = schedule_matrix

    def _store_actual_charging_rates(self):
        """ Store actual charging rates from the network in the simulator for later analysis."""
        current_rates = self.network.current_charging_rates
        agg = np.sum(current_rates)
        if self.iteration < self.charging_rates.shape[1]:
            self.charging_rates[:, self.iteration] = current_rates.T
        else:
            if not self.event_queue.empty():
                width_increase = self.event_queue.get_last_timestamp() + 1
            else:
                width_increase = self._iteration + 1
            self.charging_rates = _increase_width(self.charging_rates, width_increase)
            self.charging_rates[:, self._iteration] = current_rates.T
        self.peak = max(self.peak, agg)

    def _print(self, s):
        if self.verbose:
            print(s)

    def charging_rates_as_df(self):
        """ Return the charging rates as a pandas DataFrame, with EVSE id as columns
        and iteration as index.

        Returns:
            pandas.DataFrame: A DataFrame containing the charging rates
                of the simulation. Columns are EVSE id, and the index is
                the iteration.
        """
        return pd.DataFrame(
            data=self.charging_rates.T, columns=self.network.station_ids
        )

    def pilot_signals_as_df(self):
        """ Return the pilot signals as a pandas DataFrame

        Returns:
            pandas.DataFrame: A DataFrame containing the pilot signals
                of the simulation. Columns are EVSE id, and the index is
                the iteration.
        """
        return pd.DataFrame(data=self.pilot_signals.T, columns=self.network.station_ids)

    def index_of_evse(self, station_id):
        """ Return the numerical index of the EVSE given by station_id in the (ordered) dictionary
        of EVSEs.
        """
        if station_id not in self.network.station_ids:
            raise KeyError("EVSE {0} not found in network.".format(station_id))
        return self.network.station_ids.index(station_id)

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Implements BaseSimObj._to_dict. Certain simulator attributes are
        not serialized completely as they are not ACN-Sim objects
        (signals and scheduler exist in their own modules).

        If the Python version used is less than 3.7, datetimes cannot be
        accurately loaded. As such, a warning is thrown when the start
        attribute is serialized.

        The signals attribute is only serialized if it is natively
        JSON Serializable, otherwise None is stored.

        Only the scheduler's name is serialized.
        """
        attribute_dict = {}

        # noinspection PyProtectedMember
        registry, context_dict = self.network._to_registry(context_dict=context_dict)
        attribute_dict["network"] = registry["id"]

        registry, context_dict = self.event_queue._to_registry(
            context_dict=context_dict
        )
        attribute_dict["event_queue"] = registry["id"]

        attribute_dict["scheduler"] = (
            f"{self.scheduler.__module__}." f"{self.scheduler.__class__.__name__}"
        )

        attribute_dict["start"] = self.start.strftime("%H:%M:%S.%f %d%m%Y")

        try:
            json.dumps(self.signals)
        except TypeError:
            warnings.warn(
                "Not serializing signals as value types"
                "are not natively JSON serializable.",
                UserWarning,
            )
            attribute_dict["signals"] = None
        else:
            attribute_dict["signals"] = self.signals

        # Serialize non-nested attributes.
        nn_attr_lst = [
            "period",
            "max_recompute",
            "verbose",
            "peak",
            "_iteration",
            "_resolve",
            "_last_schedule_update",
            "schedule_history",
            "pilot_signals",
            "charging_rates",
        ]
        for attr in nn_attr_lst:
            attribute_dict[attr] = getattr(self, attr)

        ev_history = {}
        for session_id, ev in self.ev_history.items():
            # noinspection PyProtectedMember
            registry, context_dict = ev._to_registry(context_dict=context_dict)
            ev_history[session_id] = registry["id"]
        attribute_dict["ev_history"] = ev_history

        event_history = []
        for past_event in self.event_history:
            # noinspection PyProtectedMember
            registry, context_dict = past_event._to_registry(context_dict=context_dict)
            event_history.append(registry["id"])
        attribute_dict["event_history"] = event_history

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """
        Implements BaseSimObj._from_dict. Certain simulator attributes
        are not loaded completely as they are not ACN-Sim objects
        (signals and scheduler exist in their own modules).

        If the Python version used is less than 3.7, the start attribute
        is stored in ISO format instead of datetime, and a warning is
        thrown.

        The signals attribute is only loaded if it was natively
        JSON Serializable, in the original object, otherwise None is
        set as the signals attribute. The Simulator's signals can be set
        after the Simulator is loaded.

        The scheduler attribute is only accurate if the scheduler's
        constructor takes no arguments, otherwise BaseAlgorithm is
        stored. The Simulator provides a method to set the scheduler
        after the Simulator is loaded.

        """
        # noinspection PyProtectedMember
        network, loaded_dict = BaseSimObj._build_from_id(
            attribute_dict["network"], context_dict, loaded_dict=loaded_dict
        )

        # noinspection PyProtectedMember
        events, loaded_dict = BaseSimObj._build_from_id(
            attribute_dict["event_queue"], context_dict, loaded_dict=loaded_dict
        )

        scheduler_cls = locate(attribute_dict["scheduler"])
        try:
            scheduler = scheduler_cls()
        except TypeError:
            warnings.warn(
                f"Scheduler {attribute_dict['scheduler']} "
                f"requires constructor inputs. Setting "
                f"scheduler to BaseAlgorithm instead."
            )
            scheduler = BaseAlgorithm()

        start = datetime.strptime(attribute_dict["start"], "%H:%M:%S.%f %d%m%Y")

        out_obj = cls(
            network,
            scheduler,
            events,
            start,
            period=attribute_dict["period"],
            signals=attribute_dict["signals"],
            verbose=attribute_dict["verbose"],
        )
        scheduler.register_interface(Interface(out_obj))

        attr_lst = [
            "max_recompute",
            "peak",
            "_iteration",
            "_resolve",
            "_last_schedule_update",
        ]
        for attr in attr_lst:
            setattr(out_obj, attr, attribute_dict[attr])

        if attribute_dict["schedule_history"] is not None:
            out_obj.schedule_history = {
                int(key): value
                for key, value in attribute_dict["schedule_history"].items()
            }
        else:
            out_obj.schedule_history = None

        out_obj.pilot_signals = np.array(attribute_dict["pilot_signals"])
        out_obj.charging_rates = np.array(attribute_dict["charging_rates"])

        ev_history = {}
        for session_id, ev in attribute_dict["ev_history"].items():
            # noinspection PyProtectedMember
            ev_elt, loaded_dict = BaseSimObj._build_from_id(
                ev, context_dict, loaded_dict=loaded_dict
            )
            ev_history[session_id] = ev_elt
        out_obj.ev_history = ev_history

        event_history = []
        for past_event in attribute_dict["event_history"]:
            # noinspection PyProtectedMember
            loaded_event, loaded_dict = BaseSimObj._build_from_id(
                past_event, context_dict, loaded_dict=loaded_dict
            )
            event_history.append(loaded_event)
        out_obj.event_history = event_history

        return out_obj, loaded_dict

    def update_scheduler(self, new_scheduler):
        """ Updates a Simulator's schedule. """
        self.scheduler = new_scheduler
        self.scheduler.register_interface(Interface(self))
        self.max_recompute = new_scheduler.max_recompute


def _increase_width(a, target_width):
    """ Returns a new 2-D numpy array with target_width number of columns, with the contents
    of a up to the first a.shape[1] columns and 0's thereafter.

    Args:
        a (numpy.Array): 2-D numpy array to be expanded.
        target_width (int): desired number of columns; must be greater than number of columns in a
    Returns:
        numpy.Array
    """
    if target_width <= a.shape[1]:
        return a
    new_matrix = np.zeros((a.shape[0], target_width))
    new_matrix[:, : a.shape[1]] = a
    return new_matrix



from typing import Optional

import requests
from datetime import datetime
from .utils import parse_dates, http_date


class DataClient(object):
    """ API client for acndata.

    Args:
        api_token (str): API token needed to access the acndata API.
        url (str): Base url for all API calls. Defaults to the standard acndata API url.

    Attributes:
        token (str): See api_token in Args.
        url (str): See url in Args.

    """

    def __init__(self, api_token, url="https://ev.caltech.edu/api/v1/"):
        self.token = api_token
        self.url = url

    def get_sessions(self, site, cond=None, project=None, sort=None, timeseries=False):
        """ Generator to return sessions from the acndata dataset one at a time.

        Args:
            site (str): ACN ID from which data should be gathered.
            cond (str): String of conditions. See API reference for the where parameter.

        Yields:
            Dict: Session as a dictionary.

        Raises:
            ValueError: Raised if the site name is not valid.
        """
        # TODO: (zach) Use API to autodetect if a new site is added.
        if site not in {"caltech", "jpl", "office001"}:
            raise ValueError(
                "Invalid site name. Must be either 'caltech', 'jpl', or 'office001'."
            )

        limit = 100
        endpoint = "sessions/" + site
        if timeseries:
            endpoint += "/ts/"
            limit = 1
        args = []
        if cond is not None:
            args.append("where={0}".format(cond))
        if project is not None:
            args.append("project={0}".format(project))
        if sort is not None:
            args.append("sort={0}".format(sort))
        args.append("max_results={0}".format(limit))
        query_string = "?" + "&".join(args) if len(args) > 0 else ""
        r = requests.get(self.url + endpoint + query_string, auth=(self.token, ""))
        payload = r.json()
        while True:
            for s in payload["_items"]:
                parse_dates(s)
                yield s
            if "next" in payload["_links"]:
                r = requests.get(
                    self.url + payload["_links"]["next"]["href"], auth=(self.token, "")
                )
                payload = r.json()
            else:
                break

    def count_sessions(self, site, cond=None):
        """ Return the number of sessions which match the given query

        Args:
            site (str): ACN ID from which data should be gathered.
            cond (str): String of conditions. See API reference for the where parameter.

        Returns:
            int: Number of sessions which match the query.

        Raises:
            ValueError: Raised if the site name is not valid.
        """
        if site not in {"caltech", "jpl", "office001"}:
            raise ValueError("Invalid site name. Must be either 'caltech' or 'jpl'")

        endpoint = "sessions/" + site
        args = []
        if cond is not None:
            args.append("where={0}".format(cond))
        args.append("limit=1")
        query_string = "?" + "&".join(args) if len(args) > 0 else ""
        auth_header = {"Authorization": "Bearer {0}".format(self.token)}
        r = requests.head(self.url + endpoint + query_string, headers=auth_header)
        return r.headers["x-total-count"]

    def get_sessions_by_time(
        self,
        site,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        min_energy=None,
        timeseries=False,
        count=False,
    ):
        """ Wrapper for get_sessions with condition based on start and end times and a minimum energy delivered.

        Args:
            site (str): Site where data should be gathered.
            start (datetime): Only return sessions which began after start.
            end (datetime): Only return session which began before end.
            min_energy (float): Only return sessions where the kWhDelivered is greater than or equal to min_energy.
            timeseries (bool): If True return the time-series of charging rates and pilot signals. Default False.
            count (bool): If True return the number of sessions which would be returned by the function. Default False.

        Yields:
            If count is False see get_sessions else see count_sessions.

        Raises:
            See get_sessions/count_sessions.
        """
        cond = []
        if start is not None:
            cond.append('connectionTime >= "{0}"'.format(http_date(start)))
        if end is not None:
            cond.append('connectionTime <= "{0}"'.format(http_date(end)))
        if min_energy is not None:
            cond.append("kWhDelivered > {0}".format(min_energy))
        condition = " and ".join(cond)
        if count:
            return self.count_sessions(site, condition)
        else:
            return self.get_sessions(
                site, condition, sort="connectionTime", timeseries=timeseries
            )



# coding=utf-8
"""
Preprocessing functions for scheduling algorithms.
"""
from typing import List
import numpy as np

from acnportal.acnsim.interface import SessionInfo, InfrastructureInfo
from .upper_bound_estimator import UpperBoundEstimatorBase
from .utils import infrastructure_constraints_feasible, remaining_amp_periods


def enforce_pilot_limit(
    active_sessions: List[SessionInfo], infrastructure: InfrastructureInfo
) -> List[SessionInfo]:
    """ Update the max_rates vector for each session to be less than the max
        pilot supported by its EVSE.

    Args:
        active_sessions (List[SessionInfo]): List of SessionInfo objects for
            all active charging sessions.
        infrastructure (InfrastructureInfo): Description of the charging
            infrastructure.

    Returns:
        List[SessionInfo]: Active sessions with max_rates updated to be at
            most the max_pilot of the corresponding EVSE.
    """
    for session in active_sessions:
        i = infrastructure.get_station_index(session.station_id)
        session.max_rates = np.minimum(session.max_rates, infrastructure.max_pilot[i])
    return active_sessions


def reconcile_max_and_min(session: SessionInfo, choose_min: bool = True) -> SessionInfo:
    """ Modify session.max_rates[t] to equal session.min_rates[t] for times
        when max_rates[t] < min_rates[t]

    Args:
        session (SessionInfo): Session object.
        choose_min (bool): If True, when in conflict defer to the minimum
            rate. If False, defer to maximum.

    Returns:
        SessionInfo: session modified such that max_rates[t] is never less
            than min_rates[t]
    """
    mask = session.max_rates < session.min_rates
    if choose_min:
        session.max_rates[mask] = session.min_rates[mask]
    else:
        session.min_rates[mask] = session.max_rates[mask]
    return session


def expand_max_min_rates(active_sessions: List[SessionInfo]) -> List[SessionInfo]:
    """ Expand max_rates and min_rates to vectors if they are scalars. Doing so is
    helpful for scheduling algorithms that use a Model Predictive Control framework
    such as CVXPY.

    Args:
        active_sessions (List[SessionInfo]): List of SessionInfo objects for
            all active charging sessions.

    Returns:
        List[SessionInfo]: Active sessions with max_rates and min_rates
            expanded into vectors of length remaining_time.
    """
    for session in active_sessions:
        if np.isscalar(session.max_rates):
            session.max_rates = np.full(session.remaining_time, session.max_rates)
        if np.isscalar(session.min_rates):
            session.min_rates = np.full(session.remaining_time, session.min_rates)
    return active_sessions


def apply_upper_bound_estimate(
    ub_estimator: UpperBoundEstimatorBase, active_sessions: List[SessionInfo]
) -> List[SessionInfo]:
    """ Update max_rate in each SessionInfo object.

    If rampdown max_rate is less than min_rate, max_rate is set equal to min_rate.

    Args:
        ub_estimator (UpperBoundEstimatorBase): UpperBoundEstimatorBase-like
            object which estimates an upper bound on the charging rate of
            each EV based on historical data.
        active_sessions (List[SessionInfo]): List of SessionInfo objects for
            all active charging sessions.

    Returns:
        List[SessionInfo]: Active sessions with updated max_rate using rampdown.
    """
    new_sessions = expand_max_min_rates(active_sessions)
    upper_bounds = ub_estimator.get_maximum_rates(active_sessions)
    for j, session in enumerate(new_sessions):
        session.max_rates = np.minimum(
            session.max_rates, upper_bounds.get(session.station_id, float("inf"))
        )
        new_sessions[j] = reconcile_max_and_min(session)
        if np.any(session.max_rates < 32):
            pass
    return new_sessions


def apply_minimum_charging_rate(
    active_sessions: List[SessionInfo],
    infrastructure: InfrastructureInfo,
    period: int,
    override: float = float("inf"),
) -> List[SessionInfo]:
    """ Modify active_sessions so that min_rates[0] is equal to the greater of
        the session minimum rate and the EVSE minimum pilot. Sessions have their min
        rates applied in order of remaining time; i.e., sessions with less time
        remaining are allocated their min rates first.

    Args:
        active_sessions (List[SessionInfo]): List of SessionInfo objects for
            all active charging sessions.
        infrastructure (InfrastructureInfo): Description of the charging
            infrastructure.
        period (int): Length of each time period in minutes.
        override (float): Alternative minimum pilot which overrides the EVSE
            minimum if the EVSE minimum is greater than override.

    Returns:
        List[SessionInfo]: Active sessions with updated minimum charging rate
            for the first control period.
    """
    session_queue = sorted(active_sessions, key=lambda x: x.remaining_time)
    session_queue = expand_max_min_rates(session_queue)
    rates = np.zeros(len(infrastructure.station_ids))
    for j, session in enumerate(session_queue):
        i = infrastructure.station_ids.index(session.station_id)
        rates[i] = min(infrastructure.min_pilot[i], override)
        if rates[i] <= remaining_amp_periods(
            session, infrastructure, period
        ) and infrastructure_constraints_feasible(rates, infrastructure):
            # Preserve existing min_rate if it is greater than the new one
            session.min_rates[0] = max(rates[i], session.min_rates[0])
            # Increase the maximum rate if it is less than the new min.
            session_queue[j] = reconcile_max_and_min(session)
            # Keep this session as active
        else:
            # If an EV cannot be charged at the minimum rate, it should not be charged
            # in a solution to this problem. So, its max and min rates are set to 0.
            rates[i] = 0
            session.min_rates[0] = 0
            session.max_rates[0] = 0
    return session_queue


def remove_finished_sessions(
    active_sessions: List[SessionInfo],
    infrastructure: InfrastructureInfo,
    period: float,
) -> List[SessionInfo]:
    """ Remove any sessions where the remaining demand is less than threshold.
    Here, the threshold is defined as the amount of energy delivered by charging at
    the min_pilot of a session's station, at the station's voltage, for one simulation
    period.

    Args:
        active_sessions (List[SessionInfo]): List of SessionInfo objects for
            all active charging sessions.
        infrastructure (InfrastructureInfo): Description of the charging
            infrastructure.
        period (float): Length of each time period in minutes.


    Returns:
        List[SessionInfo]: Active sessions without any sessions that are finished.

    """
    modified_sessions = []
    for s in active_sessions:
        station_index = infrastructure.get_station_index(s.station_id)
        threshold = (
            infrastructure.min_pilot[station_index]
            * infrastructure.voltages[station_index]
            / (60 / period)
            / 1000
        )  # kWh
        if s.remaining_demand > threshold:
            modified_sessions.append(s)
    return modified_sessions



# coding=utf-8
""" Tests for Postprocessing algorithms. """
from unittest import TestCase
from unittest.mock import Mock

import numpy as np

from acnportal.acnsim.simulator import InvalidScheduleError
from acnportal.algorithms.postprocessing import format_array_schedule


class TestFormatArraySchedule(TestCase):
    def setUp(self) -> None:
        """ Tests for postprocessing.format_array_schedule. """
        self.infrastructure = Mock()
        self.infrastructure.station_ids = ["PS-001", "PS-002", "PS-003", "PS-004"]

    def test_invalid_array_schedule(self) -> None:
        schedule: np.ndarray = np.array([1, 2, 3])
        with self.assertRaises(InvalidScheduleError):
            _ = format_array_schedule(schedule, self.infrastructure)

    def test_scalar_array_schedule(self) -> None:
        schedule: np.ndarray = np.array([1, 2, 3, 4])
        self.assertEqual(
            format_array_schedule(schedule, self.infrastructure),
            {"PS-001": [1.0], "PS-002": [2.0], "PS-003": [3.0], "PS-004": [4.0]},
        )

    def test_array_of_arrays_schedule(self) -> None:
        schedule: np.ndarray = np.array([[1, 2], [2, 3], [3, 4], [4, 5]])
        self.assertEqual(
            format_array_schedule(schedule, self.infrastructure),
            {
                "PS-001": [1.0, 2.0],
                "PS-002": [2.0, 3.0],
                "PS-003": [3.0, 4.0],
                "PS-004": [4.0, 5.0],
            },
        )
